@@ -10,6 +10,11 @@
 #include "kwalletd_debug.h"
 
 #include "kbetterthankdialog.h"
+#include "kwalletfreedesktopcollection.h"
+#include "kwalletfreedesktopitem.h"
+#include "kwalletfreedesktopprompt.h"
+#include "kwalletfreedesktopservice.h"
+#include "kwalletfreedesktopsession.h"
 #include "kwalletwizard.h"
 
 #ifdef HAVE_GPGMEPP
@@ -26,11 +31,17 @@
 #include <KNotification>
 #include <KPasswordDialog>
 #include <KPluginFactory>
-#include <KPluginLoader>
 #include <KSharedConfig>
-#include <KToolInvocation>
 #include <kwalletentry.h>
 #include <kwindowsystem.h>
+
+#if !defined(Q_OS_WIN) && !defined(Q_OS_MAC)
+#define HAVE_X11 1
+#include <KX11Extras>
+#else
+#define HAVE_X11 0
+#endif
+
 #ifdef HAVE_GPGMEPP
 #include <gpgme++/key.h>
 #endif
@@ -44,6 +55,17 @@
 
 #include "kwalletadaptor.h"
 
+static void startManagerForKwalletd()
+{
+    if (!QStandardPaths::findExecutable(QStringLiteral("kstart")).isEmpty()) {
+        QProcess::startDetached(QStringLiteral("kstart"), {QStringLiteral("kwalletmanager5"), QStringLiteral("--"), QStringLiteral("--kwalletd")});
+    } else if (!QStandardPaths::findExecutable(QStringLiteral("kstart5")).isEmpty()) {
+        QProcess::startDetached(QStringLiteral("kstart"), {QStringLiteral("kwalletmanager5"), QStringLiteral("--"), QStringLiteral("--kwalletd")});
+    } else {
+        QProcess::startDetached(QStringLiteral("kwalletmanager5"), QStringList{QStringLiteral("--kwalletd")});
+    }
+}
+
 class KWalletTransaction
 {
 public:
@@ -60,6 +82,11 @@ public:
         }
     }
 
+    static int getTransactionId()
+    {
+        return nextTransactionId;
+    }
+
     ~KWalletTransaction()
     {
     }
@@ -105,13 +132,23 @@ KWalletD::KWalletD()
     _showingFailureNotify = false;
     _closeIdle = false;
     _idleTime = 0;
-    connect(&_closeTimers, SIGNAL(timedOut(int)), this, SLOT(timedOutClose(int)));
-    connect(&_syncTimers, SIGNAL(timedOut(int)), this, SLOT(timedOutSync(int)));
+    connect(&_closeTimers, &KTimeout::timedOut, this, &KWalletD::timedOutClose);
+    connect(&_syncTimers, &KTimeout::timedOut, this, &KWalletD::timedOutSync);
 
-    (void)new KWalletAdaptor(this);
-    // register services
-    QDBusConnection::sessionBus().registerService(QStringLiteral("org.kde.kwalletd5"));
-    QDBusConnection::sessionBus().registerObject(QStringLiteral("/modules/kwalletd5"), this);
+    KConfig kwalletrc(QStringLiteral("kwalletrc"));
+    KConfigGroup cfgWallet(&kwalletrc, "Wallet");
+    KConfigGroup cfgSecrets(&kwalletrc, "org.freedesktop.secrets");
+
+    if (cfgWallet.readEntry<bool>("apiEnabled", true)) {
+        (void)new KWalletAdaptor(this);
+
+        // register services
+        QDBusConnection::sessionBus().registerService(QStringLiteral("org.kde.kwalletd6"));
+        QDBusConnection::sessionBus().registerObject(QStringLiteral("/modules/kwalletd6"), this);
+        // register also with the KF5 names for backward compatibility
+        QDBusConnection::sessionBus().interface()->registerService(QStringLiteral("org.kde.kwalletd5"), QDBusConnectionInterface::QueueService);
+        QDBusConnection::sessionBus().registerObject(QStringLiteral("/modules/kwalletd5"), this);
+    }
 
 #ifdef Q_WS_X11
     screensaver = 0;
@@ -125,24 +162,15 @@ KWalletD::KWalletD()
     _dw->addDir(KWallet::Backend::getSaveLocation());
 
     _dw->startScan(true);
-    connect(_dw, SIGNAL(dirty(QString)), this, SLOT(emitWalletListDirty()));
+    connect(_dw, &KDirWatch::dirty, this, &KWalletD::emitWalletListDirty);
     connect(_dw, &KDirWatch::deleted, this, &KWalletD::emitWalletListDirty);
 
     _serviceWatcher.setWatchMode(QDBusServiceWatcher::WatchForOwnerChange);
-    connect(&_serviceWatcher, SIGNAL(serviceOwnerChanged(QString, QString, QString)), this, SLOT(slotServiceOwnerChanged(QString, QString, QString)));
-}
+    connect(&_serviceWatcher, &QDBusServiceWatcher::serviceOwnerChanged, this, &KWalletD::slotServiceOwnerChanged);
 
-void KWalletD::registerKWalletd4Service()
-{
-    auto bus = QDBusConnection::sessionBus().interface();
-    auto reply = bus->registerService(QStringLiteral("org.kde.kwalletd"), QDBusConnectionInterface::QueueService);
-    if (reply.isValid() && (reply.value() == QDBusConnectionInterface::ServiceQueued)) {
-        QDBusInterface _kde_kwalletd4(QStringLiteral("org.kde.kwalletd"), QStringLiteral("/MainApplication"), QStringLiteral("org.kde.KApplication"));
-        if (_kde_kwalletd4.isValid()) {
-            auto qreply = _kde_kwalletd4.call(QStringLiteral("quit"));
-        }
+    if (cfgSecrets.readEntry<bool>("apiEnabled", true)) {
+        _fdoService.reset(new KWalletFreedesktopService(this));
     }
-    QDBusConnection::sessionBus().registerObject(QStringLiteral("/modules/kwalletd"), this);
 }
 
 KWalletD::~KWalletD()
@@ -155,6 +183,16 @@ KWalletD::~KWalletD()
     qDeleteAll(_transactions);
 }
 
+QString KWalletD::encodeWalletName(const QString &name)
+{
+    return KWallet::Backend::encodeWalletName(name);
+}
+
+QString KWalletD::decodeWalletName(const QString &mangledName)
+{
+    return KWallet::Backend::decodeWalletName(mangledName);
+}
+
 #ifdef Q_WS_X11
 void KWalletD::connectToScreenSaver()
 {
@@ -194,7 +232,6 @@ QPair<int, KWallet::Backend *> KWalletD::findWallet(const QString &walletName) c
     return qMakePair(-1, static_cast<KWallet::Backend *>(nullptr));
 }
 
-static const QRegularExpression walletRegex(QStringLiteral("^[\\w\\^\\&\\'\\@\\{\\}\\[\\]\\,\\$\\=\\!\\-\\#\\(\\)\\%\\.\\+\\_\\s]+$"));
 bool KWalletD::_processing = false;
 
 void KWalletD::processTransactions()
@@ -300,10 +337,6 @@ int KWalletD::open(const QString &wallet, qlonglong wId, const QString &appid)
         return -1;
     }
 
-    if (!walletRegex.match(wallet).hasMatch()) {
-        return -1;
-    }
-
     KWalletTransaction *xact = new KWalletTransaction(connection());
     _transactions.append(xact);
 
@@ -324,17 +357,23 @@ int KWalletD::open(const QString &wallet, qlonglong wId, const QString &appid)
     return 0;
 }
 
-int KWalletD::openAsync(const QString &wallet, qlonglong wId, const QString &appid, bool handleSession)
+int KWalletD::nextTransactionId() const
+{
+    return KWalletTransaction::getTransactionId();
+}
+
+int KWalletD::openAsync(const QString &wallet,
+                        qlonglong wId,
+                        const QString &appid,
+                        bool handleSession,
+                        const QDBusConnection &connection,
+                        const QDBusMessage &message)
 {
     if (!_enabled) { // guard
         return -1;
     }
 
-    if (!walletRegex.match(wallet).hasMatch()) {
-        return -1;
-    }
-
-    KWalletTransaction *xact = new KWalletTransaction(connection());
+    KWalletTransaction *xact = new KWalletTransaction(connection);
     _transactions.append(xact);
 
     xact->appid = appid;
@@ -344,10 +383,10 @@ int KWalletD::openAsync(const QString &wallet, qlonglong wId, const QString &app
     xact->tType = KWalletTransaction::Open;
     xact->isPath = false;
     if (handleSession) {
-        qCDebug(KWALLETD_LOG) << "openAsync for " << message().service();
-        _serviceWatcher.setConnection(connection());
-        _serviceWatcher.addWatchedService(message().service());
-        xact->service = message().service();
+        qCDebug(KWALLETD_LOG) << "openAsync for " << message.service();
+        _serviceWatcher.setConnection(connection);
+        _serviceWatcher.addWatchedService(message.service());
+        xact->service = message.service();
     }
     QTimer::singleShot(0, this, SLOT(processTransactions()));
     checkActiveDialog();
@@ -355,9 +394,14 @@ int KWalletD::openAsync(const QString &wallet, qlonglong wId, const QString &app
     return xact->tId;
 }
 
+int KWalletD::openAsync(const QString &wallet, qlonglong wId, const QString &appid, bool handleSession)
+{
+    return openAsync(wallet, wId, appid, handleSession, connection(), message());
+}
+
 int KWalletD::openPathAsync(const QString &path, qlonglong wId, const QString &appid, bool handleSession)
 {
-    if (!_enabled) { // gaurd
+    if (!_enabled) { // guard
         return -1;
     }
 
@@ -401,11 +445,16 @@ void KWalletD::setupDialog(QWidget *dialog, WId wId, const QString &appid, bool
         // KF5 FIXME what should we use now instead of this:
         //         kapp->updateUserTimestamp();
     }
-    if (modal) {
-        KWindowSystem::setState(dialog->winId(), NET::Modal);
-    } else {
-        KWindowSystem::clearState(dialog->winId(), NET::Modal);
+
+#if HAVE_X11
+    if (KWindowSystem::isPlatformX11()) {
+        if (modal) {
+            KX11Extras::setState(dialog->winId(), NET::Modal);
+        } else {
+            KX11Extras::clearState(dialog->winId(), NET::Modal);
+        }
     }
+#endif
     activeDialog = dialog;
 }
 
@@ -431,11 +480,14 @@ void KWalletD::checkActiveDialog()
 
     activeDialog->show();
 
-    WId window = activeDialog->winId();
-    KWindowSystem::setState(window, NET::KeepAbove);
-    KWindowSystem::setOnAllDesktops(window, true);
-    KWindowSystem::forceActiveWindow(window);
-    KWindowSystem::raiseWindow(window);
+#if HAVE_X11
+    if (KWindowSystem::isPlatformX11()) {
+        WId window = activeDialog->winId();
+        KX11Extras::setState(window, NET::KeepAbove);
+        KX11Extras::setOnAllDesktops(window, true);
+        KX11Extras::forceActiveWindow(window);
+    }
+#endif
 }
 
 int KWalletD::doTransactionOpen(const QString &appid, const QString &wallet, bool isPath, qlonglong wId, bool modal, const QString &service)
@@ -585,28 +637,33 @@ int KWalletD::internalOpen(const QString &appid, const QString &wallet, bool isP
                     //              i18n( "&Open" ), "wallet-open"));
                     kpd->setWindowTitle(i18n("KDE Wallet Service"));
                     kpd->setIcon(QIcon::fromTheme(QStringLiteral("kwalletmanager")));
-                    if (w != KWindowSystem::activeWindow() && w != 0L) {
+
+#if HAVE_X11
+                    if (KWindowSystem::isPlatformX11() && w != KX11Extras::activeWindow() && w != 0L) {
                         // If the dialog is modal to a minimized window it
                         // might not be visible
                         // (but still blocking the calling application).
                         // Notify the user about
                         // the request to open the wallet.
                         KNotification *notification =
-                            new KNotification(QStringLiteral("needsPassword"), KNotification::Persistent | KNotification::CloseWhenWidgetActivated);
-                        notification->setWidget(kpd);
-                        QStringList actions;
+                            new KNotification(QStringLiteral("needsPassword"), KNotification::Persistent | KNotification::CloseWhenWindowActivated);
+                        notification->setWindow(kpd->windowHandle());
+                        QString actionText;
                         if (appid.isEmpty()) {
                             notification->setText(i18n("An application has requested to open a wallet (%1).", wallet.toHtmlEscaped()));
-                            actions.append(i18nc("Text of a button for switching to the (unnamed) application requesting a password", "Switch there"));
+                            actionText = i18nc("Text of a button for switching to the (unnamed) application requesting a password", "Switch there");
                         } else {
                             notification->setText(i18n("<b>%1</b> has requested to open a wallet (%2).", appid.toHtmlEscaped(), wallet.toHtmlEscaped()));
-                            actions.append(
-                                i18nc("Text of a button for switching to the application requesting a password", "Switch to %1", appid.toHtmlEscaped()));
+                            actionText =
+                                i18nc("Text of a button for switching to the application requesting a password", "Switch to %1", appid.toHtmlEscaped());
                         }
-                        notification->setActions(actions);
-                        connect(notification, SIGNAL(action1Activated()), this, SLOT(activatePasswordDialog()));
+
+                        KNotificationAction *action = notification->addAction(actionText);
+                        connect(action, &KNotificationAction::activated, this, &KWalletD::activatePasswordDialog);
                         notification->sendEvent();
                     }
+#endif
+
                     while (!b->isOpen()) {
                         setupDialog(kpd, w, appid, modal);
                         if (kpd->exec() == QDialog::Accepted) {
@@ -740,7 +797,7 @@ int KWalletD::internalOpen(const QString &appid, const QString &wallet, bool isP
         }
         Q_EMIT walletOpened(wallet);
         if (_wallets.count() == 1 && _launchManager) {
-            KToolInvocation::startServiceByDesktopName(QStringLiteral("kwalletmanager5-kwalletd"));
+            startManagerForKwalletd();
         }
     } else {
         // prematurely add a reference so that the wallet does not close while
@@ -838,8 +895,8 @@ bool KWalletD::isAuthorizedApp(const QString &appid, const QString &wallet, WId
 int KWalletD::deleteWallet(const QString &wallet)
 {
     int result = -1;
-    QString path = KWallet::Backend::getSaveLocation() + "/" + wallet + ".kwl";
-    QString pathSalt = KWallet::Backend::getSaveLocation() + "/" + wallet + ".salt";
+    QString path = KWallet::Backend::getSaveLocation() + "/" + encodeWalletName(wallet) + ".kwl";
+    QString pathSalt = KWallet::Backend::getSaveLocation() + "/" + encodeWalletName(wallet) + ".salt";
 
     if (QFile::exists(path)) {
         const QPair<int, KWallet::Backend *> walletInfo = findWallet(wallet);
@@ -902,7 +959,7 @@ void KWalletD::doTransactionChangePassword(const QString &appid, const QString &
     if (!w) {
         handle = doTransactionOpen(appid, wallet, false, wId, false, QLatin1String(""));
         if (-1 == handle) {
-            KMessageBox::sorryWId((WId)wId,
+            KMessageBox::errorWId((WId)wId,
                                   i18n("Unable to open wallet. The wallet must be opened in order to change the password."),
                                   i18n("KDE Wallet Service"));
             return;
@@ -938,12 +995,12 @@ void KWalletD::doTransactionChangePassword(const QString &appid, const QString &
                 w->setPassword(p.toUtf8());
                 int rc = w->close(true);
                 if (rc < 0) {
-                    KMessageBox::sorryWId((WId)wId, i18n("Error re-encrypting the wallet. Password was not changed."), i18n("KDE Wallet Service"));
+                    KMessageBox::errorWId((WId)wId, i18n("Error re-encrypting the wallet. Password was not changed."), i18n("KDE Wallet Service"));
                     reclose = true;
                 } else {
                     rc = w->open(p.toUtf8());
                     if (rc < 0) {
-                        KMessageBox::sorryWId((WId)wId, i18n("Error reopening the wallet. Data may be lost."), i18n("KDE Wallet Service"));
+                        KMessageBox::errorWId((WId)wId, i18n("Error reopening the wallet. Data may be lost."), i18n("KDE Wallet Service"));
                         reclose = true;
                     }
                 }
@@ -992,14 +1049,14 @@ int KWalletD::internalClose(KWallet::Backend *const w, const int handle, const b
     return -1;
 }
 
-int KWalletD::close(int handle, bool force, const QString &appid)
+int KWalletD::close(int handle, bool force, const QString &appid, const QDBusMessage &message)
 {
     KWallet::Backend *w = _wallets.value(handle);
 
     if (w) {
         if (_sessions.hasSession(appid, handle)) {
             // remove one handle for the application
-            bool removed = _sessions.removeSession(appid, message().service(), handle);
+            bool removed = _sessions.removeSession(appid, message.service(), handle);
             // alternatively try sessionless
             if (removed || _sessions.removeSession(appid, QLatin1String(""), handle)) {
                 w->deref();
@@ -1011,6 +1068,11 @@ int KWalletD::close(int handle, bool force, const QString &appid)
     return -1; // not open to begin with, or other error
 }
 
+int KWalletD::close(int handle, bool force, const QString &appid)
+{
+    return close(handle, force, appid, message());
+}
+
 bool KWalletD::isOpen(const QString &wallet)
 {
     const QPair<int, KWallet::Backend *> walletInfo = findWallet(wallet);
@@ -1049,7 +1111,7 @@ QStringList KWalletD::wallets() const
         if (fn.endsWith(QLatin1String(".kwl"))) {
             fn.truncate(fn.length() - 4);
         }
-        rc += fn;
+        rc += decodeWalletName(fn);
     }
     return rc;
 }
@@ -1329,6 +1391,27 @@ int KWalletD::writeMap(int handle, const QString &folder, const QString &key, co
         b->writeEntry(&e);
         initiateSync(handle);
         emitFolderUpdated(b->walletName(), folder);
+        emitEntryUpdated(b->walletName(), folder, key);
+        return 0;
+    }
+
+    return -1;
+}
+
+int KWalletD::writeEntry(int handle, const QString &folder, const QString &key, const QByteArray &value, const QString &appid)
+{
+    KWallet::Backend *b;
+
+    if ((b = getWallet(appid, handle))) {
+        b->setFolder(folder);
+        KWallet::Entry e;
+        e.setKey(key);
+        e.setValue(value);
+        e.setType(KWallet::Wallet::Stream);
+        b->writeEntry(&e);
+        initiateSync(handle);
+        emitFolderUpdated(b->walletName(), folder);
+        emitEntryUpdated(b->walletName(), folder, key);
         return 0;
     }
 
@@ -1354,25 +1437,6 @@ int KWalletD::writeEntry(int handle, const QString &folder, const QString &key,
     return -1;
 }
 
-int KWalletD::writeEntry(int handle, const QString &folder, const QString &key, const QByteArray &value, const QString &appid)
-{
-    KWallet::Backend *b;
-
-    if ((b = getWallet(appid, handle))) {
-        b->setFolder(folder);
-        KWallet::Entry e;
-        e.setKey(key);
-        e.setValue(value);
-        e.setType(KWallet::Wallet::Stream);
-        b->writeEntry(&e);
-        initiateSync(handle);
-        emitFolderUpdated(b->walletName(), folder);
-        return 0;
-    }
-
-    return -1;
-}
-
 int KWalletD::writePassword(int handle, const QString &folder, const QString &key, const QString &value, const QString &appid)
 {
     KWallet::Backend *b;
@@ -1386,6 +1450,7 @@ int KWalletD::writePassword(int handle, const QString &folder, const QString &ke
         b->writeEntry(&e);
         initiateSync(handle);
         emitFolderUpdated(b->walletName(), folder);
+        emitEntryUpdated(b->walletName(), folder, key);
         return 0;
     }
 
@@ -1436,6 +1501,7 @@ int KWalletD::removeEntry(int handle, const QString &folder, const QString &key,
         bool rc = b->removeEntry(key);
         initiateSync(handle);
         emitFolderUpdated(b->walletName(), folder);
+        emitEntryDeleted(b->walletName(), folder, key);
         return rc ? 0 : -3;
     }
 
@@ -1549,12 +1615,19 @@ int KWalletD::renameEntry(int handle, const QString &folder, const QString &oldN
         int rc = b->renameEntry(oldName, newName);
         initiateSync(handle);
         emitFolderUpdated(b->walletName(), folder);
+        emitEntryRenamed(b->walletName(), folder, oldName, newName);
         return rc;
     }
 
     return -1;
 }
 
+int KWalletD::renameWallet(const QString &oldName, const QString &newName)
+{
+    const QPair<int, KWallet::Backend *> walletInfo = findWallet(oldName);
+    return walletInfo.second->renameWallet(newName);
+}
+
 QStringList KWalletD::users(const QString &wallet) const
 {
     const QPair<int, KWallet::Backend *> walletInfo = findWallet(wallet);
@@ -1587,6 +1660,21 @@ void KWalletD::emitFolderUpdated(const QString &wallet, const QString &folder)
     Q_EMIT folderUpdated(wallet, folder);
 }
 
+void KWalletD::emitEntryUpdated(const QString &wallet, const QString &folder, const QString &key)
+{
+    Q_EMIT entryUpdated(wallet, folder, key);
+}
+
+void KWalletD::emitEntryRenamed(const QString &wallet, const QString &folder, const QString &oldName, const QString &newName)
+{
+    Q_EMIT entryRenamed(wallet, folder, oldName, newName);
+}
+
+void KWalletD::emitEntryDeleted(const QString &wallet, const QString &folder, const QString &key)
+{
+    Q_EMIT entryDeleted(wallet, folder, key);
+}
+
 void KWalletD::emitWalletListDirty()
 {
     const QStringList walletsInDisk = wallets();
@@ -1778,10 +1866,6 @@ int KWalletD::pamOpen(const QString &wallet, const QByteArray &passwordHash, int
         return -1;
     }
 
-    if (!walletRegex.match(wallet).hasMatch()) {
-        return -1;
-    }
-
     // check if the wallet is already open
     QPair<int, KWallet::Backend *> walletInfo = findWallet(wallet);
     int rc = walletInfo.first;
@@ -1824,9 +1908,12 @@ int KWalletD::pamOpen(const QString &wallet, const QByteArray &passwordHash, int
     Q_EMIT walletOpened(wallet);
 
     if (_wallets.count() == 1 && _launchManager) {
-        KToolInvocation::startServiceByDesktopName(QStringLiteral("kwalletmanager5-kwalletd"));
+        startManagerForKwalletd();
     }
 
     return handle;
 }
+
 // vim: tw=220:ts=4
+
+#include "moc_kwalletd.cpp"
